#include "elligator.h"

static mpz_class invert(mpz_class denom,mpz_class p)
{
  mpz_class gcd,recip,other;
  mpz_gcdext(gcd.get_mpz_t(),recip.get_mpz_t(),other.get_mpz_t(),denom.get_mpz_t(),p.get_mpz_t());
  /* gcd = recip denom + other p */
  recip %= p;
  recip += p;
  recip %= p;
  assert(gcd == 1);
  assert((recip*denom)%p == 1);
  assert(recip > 0);
  assert(recip < p);
  return recip;
}

point_affine elligator(curve C,mpz_class r,bit twist,mpz_class base)
{
  mpz_class p = C.parent();

  assert(r >= 2);
  assert(r*2 <= p-1);
  mpz_class recip = invert(r*r-1,p);

  natural A = C.montgomeryA();
  bit trouble = A.isequal(0);

  natural v = (A*natural(recip))%p;
  v = v.mux(natural(base),trouble);
  natural x = (v+A).minimod(p);
  natural m = (x*v)%p;
  m = ((m+1)*v)%p;
  bit nonsquare = m.powmod((p-1)/2,p).bitatpos(1);
  x = natural(0).subtractmod(x,p);
  x = v.mux(x,nonsquare^twist);

  return point_affine(C,x);
}

mpz_class elligator_base(mpz_class p,std::vector<mpz_class> factors)
{
  for (mpz_class x = 1;;++x) {
    mpz_class m = (x*x*x+x) % p;
    mpz_class exponent = (p+1)/4;
    mpz_class y;
    mpz_powm(y.get_mpz_t(),m.get_mpz_t(),exponent.get_mpz_t(),p.get_mpz_t());
    if ((y*y) % p != m) continue;

    int ok = 1;
    for (long long i = 0;i < factors.size();++i) {
      if (factors[i] == 2) continue;
      exponent = 1;
      for (long long j = 0;j < factors.size();++j)
        if (j != i)
          exponent *= factors[j];

      mpz_class X2 = 1;
      mpz_class Z2 = 0;
      mpz_class X3 = x;
      mpz_class Z3 = 1;
      for (long long k = mpz_sizeinbase(exponent.get_mpz_t(),2) - 1;k >= 0;--k) {
        if (mpz_tstbit(exponent.get_mpz_t(),k)) {
          mpz_class T = X2; X2 = X3; X3 = T;
                    T = Z2; Z2 = Z3; Z3 = T;
        }
        mpz_class A = X2+Z2;
        mpz_class AA = A*A;
        mpz_class B = X2-Z2;
        mpz_class BB = B*B;
        mpz_class E = AA-BB;
        mpz_class C = X3+Z3;
        mpz_class D = X3-Z3;
        mpz_class DA = D*A;
        mpz_class CB = C*B;
        mpz_class X5 = (DA+CB)*(DA+CB);
        mpz_class Z5 = x*(DA-CB)*(DA-CB);
        mpz_class X4 = AA*BB*2;
        mpz_class Z4 = E*(BB*2+E);

        X2 = X4 % p; Z2 = Z4 % p;
        X3 = X5 % p; Z3 = Z5 % p;

        if (mpz_tstbit(exponent.get_mpz_t(),k)) {
          mpz_class T = X2; X2 = X3; X3 = T;
                    T = Z2; Z2 = Z3; Z3 = T;
        }
      }

      if (X2 == 0 || Z2 == 0) {
        ok = 0;
        break;
      }
    }

    if (ok) return x;
  }
}
